// Package masterelection lets you run a simple master-election
// protocol on top of etcd.
package masterelection

import (
	"log"
	"os"
	"sync"
	"time"

	"github.com/coreos/go-etcd/etcd"
)

// Possible values for Role.
const (
	RoleUnknown = iota
	RoleSlave
	RoleMaster
)

// How long to sleep on etcd Create() errors.
var errorDelay = 200 * time.Millisecond

// Role of a participant in the master election protocol.
type Role int

func (r Role) String() string {
	switch r {
	case RoleSlave:
		return "slave"
	case RoleMaster:
		return "master"
	}
	return "unknown"
}

// State of the master election protocol (istantaneous snapshot for a
// node). It stores the role of the node, and the data associated with
// the current master.
type State struct {
	Role       Role
	MasterData string
}

// Equal returns true if the two states are identical.
func (s State) Equal(other State) bool {
	return s.Role == other.Role && s.MasterData == other.MasterData
}

// EtcdClient is the etcd interface used by this package.
type EtcdClient interface {
	Create(key string, value string, ttl uint64) (*etcd.Response, error)
	Delete(key string, recursive bool) (*etcd.Response, error)
	Get(key string, sort, recursive bool) (*etcd.Response, error)
	CompareAndSwap(key string, value string, ttl uint64, prevValue string, prevIndex uint64) (*etcd.Response, error)
	Watch(prefix string, waitIndex uint64, recursive bool, receiver chan *etcd.Response, stop chan bool) (*etcd.Response, error)
}

// MasterElection runs a master election protocol on top of etcd,
// using an etcd object as a lock. Multiple master elections can run
// at the same time on the same etcd instance just by using separate
// lock files.
//
// Each participant is identified by a unique blob of data, which is
// opaque to the master election protocol itself. The data associated
// with the currently elected master (if any) can be retrieved at any
// time by calling GetMasterData.
//
type MasterElection struct {
	client EtcdClient
	data   string
	path   string
	ttl    uint64

	stateLock sync.Mutex
	stateCh   chan State
	state     State

	// Logger for debug messages and state changes. By default
	// it will send output to os.Stderr.
	Log *log.Logger
}

// New creates a new MasterElection object that will establish a lock
// on 'path'. It will send state transitions to 'sch', if provided. If
// 'sch' is not nil, it will be closed when Run() terminates.
//
// All participants in the protocol should use the same ttl.
func New(client EtcdClient, path, data string, ttl uint64, sch chan State) *MasterElection {
	if ttl < 1 {
		ttl = 1
	}
	return &MasterElection{
		client:  client,
		path:    path,
		data:    data,
		ttl:     ttl,
		stateCh: sch,
		Log:     log.New(os.Stderr, "masterelection: ", 0),
	}
}

// Valid returns true if the protocol has elected a master.
func (m *MasterElection) Valid() bool {
	m.stateLock.Lock()
	defer m.stateLock.Unlock()
	return m.state.Role != RoleUnknown
}

// IsMaster returns true if this participant is the elected master.
func (m *MasterElection) IsMaster() bool {
	m.stateLock.Lock()
	defer m.stateLock.Unlock()
	return m.state.Role == RoleMaster
}

// GetMasterData returns the opaque data associated with the currently
// elected master.
func (m *MasterElection) GetMasterData() string {
	m.stateLock.Lock()
	defer m.stateLock.Unlock()
	return m.state.MasterData
}

func (m *MasterElection) setState(role Role, masterData string) {
	state := State{
		Role:       role,
		MasterData: masterData,
	}
	var changed bool
	m.stateLock.Lock()
	if changed = !m.state.Equal(state); changed {
		m.Log.Printf("%s (%s) -> %s (%s)", m.state.Role.String(), m.state.MasterData, role.String(), masterData)
		m.state = state
	}
	m.stateLock.Unlock()
	if changed && m.stateCh != nil {
		m.stateCh <- state
	}
}

func (m *MasterElection) runMaster(index uint64, stop chan bool) {
	m.setState(RoleMaster, m.data)

	// If we renew the lease every TTL / N, we allow N renewal
	// errors before we stop believing being the master.
	ttl := time.Second * time.Duration(m.ttl)
	tick := time.NewTicker(ttl / 3)
	lastUpdate := time.Now()

	for {
		select {
		case t := <-tick.C:
			// To verify that we actually are still the
			// master (not just we believe we are), try a
			// compare-and-swap operation to check that
			// the stored master address is still our own,
			// and no-one stole our lock. If not, the TTL
			// will be updated (and the lock renewed).
			resp, err := m.client.CompareAndSwap(m.path, m.data, m.ttl, m.data, index)
			if err != nil {
				m.Log.Printf("error updating lock: %s", err)

				// If we can't renew the lock for a
				// TTL, we must assume we lost it.
				if t.Sub(lastUpdate) > ttl {
					m.Log.Println("too many errors, lost lock")
					return
				}
				continue
			}
			index = resp.EtcdIndex
			lastUpdate = t
		case <-stop:
			// Facilitate a master re-election by dropping
			// the lock rather than letting it expire.
			m.Log.Println("releasing masterelection lock")
			m.client.Delete(m.path, false)
			return
		}
	}
}

func (m *MasterElection) runSlave(index uint64, stop chan bool) {
	// It would be best if we could simply retrieve the master
	// lock file at the specified index. But we can only Get from
	// HEAD, so in case of a quick master transition we might
	// retrieve data at an index greater than 'index'.
	//
	// In this case, we skip updating state, since the Watch call
	// will immediately return (with the change that caused the
	// index to increase).
	resp, err := m.client.Get(m.path, false, false)
	if err == nil && resp.Node.ModifiedIndex <= index {
		m.setState(RoleSlave, resp.Node.Value)
	}

	for {
		resp, err = m.client.Watch(m.path, index+1, false, nil, stop)
		if err != nil {
			if err != etcd.ErrWatchStoppedByUser {
				m.Log.Printf("slave Watch() error: %v", err)
			}
			return
		}

		// If the lock has expired, or it has been removed,
		// try to acquire it again.
		if resp.Action == "delete" || resp.Action == "expire" {
			m.Log.Println("lock is gone")
			return
		}

		m.setState(RoleSlave, resp.Node.Value)
		index = resp.EtcdIndex
	}
}

// Run the master election protocol, until the stop channel is closed.
func (m *MasterElection) Run(stop chan bool) {
	// If an update channel is set, close it on exit.
	if m.stateCh != nil {
		defer close(m.stateCh)
	}

	for {
		// Quick non-blocking check for the stop channel.
		select {
		case <-stop:
			return
		default:
		}

		// Try to acquire the lock. This call will only succeed
		// if the lockfile does not exist (either because it
		// expired, or the previous master exited gracefully and
		// deleted it).
		response, err := m.client.Create(m.path, m.data, m.ttl)

		if err == nil {
			// Howdy, we're the master now. Wait a while
			// and renew our TTL.
			m.runMaster(response.EtcdIndex, stop)
		} else if etcdErr, ok := err.(*etcd.EtcdError); ok {
			// We're not the master. Wait until the lock
			// is deleted or expires.
			m.runSlave(etcdErr.Index, stop)
		} else {
			// An error of some other sort! Retry.
			m.Log.Printf("unexpected error: %v", err)
			time.Sleep(errorDelay)
		}
	}
}
